
/*
Vocab:
* `row` - the 0-based index of a line.
* `col` - the 0-based index of a column.

Keeping points synchronized
* when text is inserted at point A
  - point B must be moved down (if A contains newlines)
  - point B must be moved right (if A is on the same line)
* when text is deleted forward at A
  - point B is moved up (if !same line and line was joined)
  - point B is moved left (if on same line)
* when text is deleted backward at A
  - point B is moved up (if !same line and line was joined)
  - point B is moved left (if on same line)
*/

(function() {
  var Buffer, EventEmitter, Point, Region,
    __hasProp = Object.prototype.hasOwnProperty,
    __extends = function(child, parent) { for (var key in parent) { if (__hasProp.call(parent, key)) child[key] = parent[key]; } function ctor() { this.constructor = child; } ctor.prototype = parent.prototype; child.prototype = new ctor; child.__super__ = parent.prototype; return child; };

  EventEmitter = require('events').EventEmitter;

  Buffer = (function(_super) {

    __extends(Buffer, _super);

    function Buffer(text, saveCursor, placeCursor) {
      this.saveCursor = saveCursor != null ? saveCursor : null;
      this.placeCursor = placeCursor != null ? placeCursor : null;
      this.lines = [""];
      this.setText(text.replace(/\r/g, ""));
      this.undoStack = [];
      this.redoStack = [];
      this.currentSteps = [];
    }

    Buffer.prototype.point = function(row, col, anchor) {
      return new Point(this, row, col, anchor);
    };

    Buffer.prototype.text = function(row, col) {
      var _ref;
      if (row == null) row = null;
      if (col == null) col = null;
      if ((row != null) && (col != null)) {
        return (_ref = this.lines[row]) != null ? _ref[col] : void 0;
      } else if (row != null) {
        return this.lines[row];
      } else {
        return this.lines.join("\n");
      }
    };

    Buffer.prototype.lineCount = function() {
      return this.lines.length;
    };

    Buffer.prototype.lineLength = function(row) {
      var _ref;
      if (row == null) throw new Error("Buffer#lineLength() needs a row number");
      return (_ref = this.text(row)) != null ? _ref.length : void 0;
    };

    Buffer.prototype.search = function(match, startRow, startCol) {
      if (startRow == null) startRow = 0;
      if (startCol == null) startCol = 0;
      if (match instanceof RegExp) {
        return this.searchRegex(match, startRow, startCol);
      } else {
        return this.searchText(match, startRow, startCol);
      }
    };

    Buffer.prototype.searchText = function(subText, startRow, startCol) {
      var begin, end, length, maxRow, result, text;
      text = this.text(startRow).substr(startCol);
      maxRow = this.lineCount();
      while (startRow < maxRow) {
        result = text.indexOf(subText);
        if (~result) {
          length = subText.length;
          begin = this.point(startRow, startCol + result);
          end = this.point(startRow, startCol + result + length);
          return new Region(begin, end);
        }
        startCol = 0;
        text = this.text(++startRow);
      }
      return null;
    };

    Buffer.prototype.searchRegex = function(regex, startRow, startCol) {
      var begin, end, index, length, maxRow, region, result, text;
      text = this.text(startRow).substr(startCol);
      maxRow = this.lineCount();
      while (startRow < maxRow) {
        result = regex.exec(text);
        if (result) {
          length = result[0].length;
          index = result.index;
          begin = this.point(startRow, startCol + index);
          end = this.point(startRow, startCol + index + length);
          region = new Region(begin, end);
          region.captures = result;
          return region;
        }
        startCol = 0;
        text = this.text(++startRow);
      }
      return null;
    };

    Buffer.prototype.searchAll = function(match) {
      var col, region, regions, row;
      regions = [];
      row = 0;
      col = 0;
      while (region = this.search(match, row, col)) {
        regions.push(region);
        row = region.end.row;
        col = region.end.col;
      }
      return regions;
    };

    Buffer.prototype.replace = function(match, newText, startRow, startCol) {
      var capture, i, matchRegion, _len, _ref;
      if (startRow == null) startRow = 0;
      if (startCol == null) startCol = 0;
      matchRegion = this.search(match, startRow, startCol);
      if (!matchRegion) return null;
      if (matchRegion.captures) {
        _ref = matchRegion.captures;
        for (i = 0, _len = _ref.length; i < _len; i++) {
          capture = _ref[i];
          newText = newText.replace(new RegExp("[$]" + i, "g"), capture);
        }
      }
      matchRegion.replaceWith(newText);
      return matchRegion;
    };

    Buffer.prototype.replaceAll = function(match, newText, startRow, startCol) {
      var matchRegion, replacementCount;
      if (startRow == null) startRow = 0;
      if (startCol == null) startCol = 0;
      replacementCount = 0;
      while (matchRegion = this.replace(match, newText, startRow, startCol)) {
        startRow = matchRegion.end.row;
        startCol = matchRegion.end.col;
        replacementCount++;
      }
      return replacementCount;
    };

    Buffer.prototype.deleteLine = function(row) {
      var line;
      line = this.text(row);
      if (row === 0 && this.lineCount() === 1) {
        this.setLine(0, "");
      } else {
        this.lines.splice(row, 1);
        if (!this.noHistory) {
          this.currentSteps.push({
            type: "delete",
            row: row,
            oldText: line
          });
        }
        this.emit("line:delete", {
          row: row
        });
      }
      return line;
    };

    Buffer.prototype.setLine = function(row, text) {
      var oldText;
      oldText = this.text(row);
      this.lines[row] = text;
      if (!this.noHistory) {
        this.currentSteps.push({
          type: "change",
          row: row,
          oldText: oldText,
          newText: text
        });
      }
      this.emit("line:change", {
        row: row,
        text: text
      });
    };

    Buffer.prototype.insertLine = function(row, text) {
      this.lines.splice(row, 0, text);
      if (!this.noHistory) {
        this.currentSteps.push({
          type: "insert",
          row: row,
          newText: text
        });
      }
      this.emit("line:insert", {
        row: row,
        text: text
      });
    };

    Buffer.prototype.setText = function(text) {
      this.lines = text.split("\n");
      return this.emit("reset");
    };

    Buffer.prototype.undo = function() {
      var prev, step, steps, _i, _len, _ref;
      this.commitTransaction();
      steps = this.undoStack.pop();
      if (!steps) return;
      this.redoStack.push(steps);
      this.noHistory = true;
      _ref = steps.slice().reverse();
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        step = _ref[_i];
        switch (step.type) {
          case "change":
            this.setLine(step.row, step.oldText);
            break;
          case "delete":
            this.insertLine(step.row, step.oldText);
            break;
          case "insert":
            this.deleteLine(step.row);
        }
      }
      if ((prev = this.undoStack[this.undoStack.length - 1]) && prev.cursor) {
        this.placeCursor(prev.cursor);
      }
      return this.noHistory = false;
    };

    Buffer.prototype.redo = function() {
      var step, steps, _i, _len, _ref;
      steps = this.redoStack.pop();
      if (!steps) return;
      this.undoStack.push(steps);
      this.noHistory = true;
      _ref = steps.slice();
      for (_i = 0, _len = _ref.length; _i < _len; _i++) {
        step = _ref[_i];
        switch (step.type) {
          case "change":
            this.setLine(step.row, step.newText);
            break;
          case "delete":
            this.deleteLine(step.row);
            break;
          case "insert":
            this.insertLine(step.row, step.newText);
        }
      }
      if (steps.cursor) this.placeCursor(steps.cursor);
      return this.noHistory = false;
    };

    Buffer.prototype.commitTransaction = function() {
      if (!this.currentSteps.length) return;
      if (this.saveCursor) this.currentSteps.cursor = this.saveCursor();
      this.undoStack.push(this.currentSteps);
      return this.currentSteps = [];
    };

    Buffer.prototype.insert = function(text, row, col) {
      var i, insertedLineCount, insertedLines, lastLine, lastRow, line, textAfter, textBefore, _len, _ref;
      line = this.text(row);
      textBefore = line.substr(0, col);
      textAfter = line.substr(col);
      if (!~text.indexOf("\n")) {
        this.setLine(row, textBefore + text + textAfter);
        return this.point(row, textBefore.length + text.length);
      } else {
        insertedLines = text.split("\n");
        insertedLineCount = insertedLines.length;
        this.setLine(row, textBefore + insertedLines[0]);
        _ref = insertedLines.slice(1, -1);
        for (i = 0, _len = _ref.length; i < _len; i++) {
          line = _ref[i];
          this.insertLine(row + i + 1, line);
        }
        lastRow = row + insertedLineCount - 1;
        lastLine = insertedLines[insertedLineCount - 1];
        this.insertLine(lastRow, lastLine + textAfter);
        return this.point(lastRow, lastLine.length);
      }
    };

    Buffer.prototype.overwrite = function(text, row, col) {
      var line, textAfter, textBefore;
      line = this.text(row);
      textBefore = line.substr(0, col);
      textAfter = line.substr(col + text.length);
      this.setLine(row, textBefore + text + textAfter);
      return this.point(row, textBefore.length + text.length);
    };

    Buffer.prototype.insertNewLine = function(row, col) {
      var line, textAfter, textBefore;
      line = this.text(row);
      textBefore = line.substr(0, col);
      textAfter = line.substr(col);
      this.setLine(row, textBefore);
      return this.insertLine(row + 1, textAfter);
    };

    Buffer.prototype.joinLines = function(row) {
      var line, line2;
      line = this.text(row);
      line2 = this.text(row + 1);
      this.setLine(row, line + line2);
      return this.deleteLine(row + 1);
    };

    Buffer.prototype.wordAt = function(row, col, wordRe) {
      var index, isNext, length, line, match, regex, region, text, _ref;
      if (wordRe == null) wordRe = "\\w+";
      if (row instanceof Point) _ref = row, row = _ref.row, col = _ref.col;
      line = this.text(row);
      regex = new RegExp("(" + wordRe + ")|(.)", "g");
      isNext = false;
      while (match = regex.exec(line)) {
        text = match[1] || match[2];
        index = match.index;
        length = text.length;
        if (isNext || (col <= index + length)) {
          region = new Region(this.point(row, index), this.point(row, index + length));
          region.isSolid = !!match[1];
          if (!isNext && !region.isSolid && index + 1 !== line.length) {
            isNext = true;
            continue;
          }
          return region;
        }
      }
      return new Region(this.point(row, col), this.point(row, col));
    };

    Buffer.prototype.shiftLinesUp = function(beginRow, endRow) {
      var prevLine;
      if (!beginRow) return false;
      prevLine = this.deleteLine(beginRow - 1);
      this.insertLine(endRow, prevLine);
      return true;
    };

    Buffer.prototype.shiftLinesDown = function(beginRow, endRow) {
      var nextLine;
      if (endRow === this.lineCount() - 1) return false;
      nextLine = this.deleteLine(endRow + 1);
      this.insertLine(beginRow, nextLine);
      return true;
    };

    Buffer.prototype.deleteLines = function(beginRow, endRow) {
      var i;
      for (i = beginRow; beginRow <= endRow ? i <= endRow : i >= endRow; beginRow <= endRow ? i++ : i--) {
        this.deleteLine(beginRow);
      }
    };

    return Buffer;

  })(EventEmitter);

  Point = (function(_super) {

    __extends(Point, _super);

    function Point(buffer, row, col, anchor) {
      this.buffer = buffer;
      this.row = row;
      this.col = col;
      this.anchor = anchor != null ? anchor : true;
    }

    Point.prototype.moveTo = function(row, col) {
      if (row instanceof Point) {
        this.row = row.row;
        this.col = row.col;
        return this.emit("move");
      } else {
        if (row != null) this.row = row;
        if (col != null) this.col = col;
        return this.emit("move");
      }
    };

    Point.prototype.equals = function(point) {
      return this.row === point.row && this.col === point.col;
    };

    Point.prototype.isBefore = function(point) {
      var sameRow;
      sameRow = this.row === point.row;
      return this.row < point.row || (sameRow && this.col < point.col);
    };

    Point.prototype.isAfter = function(point) {
      return !this.isBefore(point) && !this.equals(point);
    };

    Point.prototype.clone = function() {
      return new Point(this.buffer, this.row, this.col, this.anchor);
    };

    Point.prototype.toString = function() {
      return "(" + this.row + ", " + this.col + ")";
    };

    Point.prototype.round = function() {
      var lastCol, lastRow, newCol, newRow;
      newRow = null;
      newCol = null;
      lastRow = this.buffer.lineCount() - 1;
      lastCol = this.buffer.lineLength(this.row);
      if (this.row < 0) newRow = 0;
      if (this.col < 0) newCol = 0;
      if (this.row > lastRow) newRow = lastRow;
      if (this.col > lastCol) newCol = lastCol;
      if ((newRow != null) || (newCol != null)) return this.moveTo(newRow, newCol);
    };

    Point.prototype.prevLoc = function() {
      if (this.col === 0) {
        if (this.row === 0) {
          return this.buffer.point(0, 0);
        } else {
          return this.buffer.point(this.row - 1, this.buffer.lineLength(this.row - 1));
        }
      } else {
        return this.buffer.point(this.row, this.col - 1);
      }
    };

    Point.prototype.nextLoc = function() {
      if (this.col === this.buffer.lineLength(this.row)) {
        if (this.row === this.buffer.lineCount() - 1) {
          return this.buffer.point(this.row, this.col);
        } else {
          return this.buffer.point(this.row + 1, 0);
        }
      } else {
        return this.buffer.point(this.row, this.col + 1);
      }
    };

    Point.prototype.moveToLineBegin = function() {
      this.idealCol = 0;
      return this.moveTo(null, 0);
    };

    Point.prototype.moveToLineEnd = function() {
      this.idealCol = 0;
      return this.moveTo(null, this.buffer.lineLength(this.row));
    };

    Point.prototype.moveLeft = function() {
      if (this.isAtDocBegin()) return;
      this.idealCol = 0;
      if (this.col === 0) {
        if (this.row) {
          return this.moveTo(this.row - 1, this.buffer.lineLength(this.row - 1));
        }
      } else {
        return this.moveTo(null, this.col - 1);
      }
    };

    Point.prototype.moveRight = function() {
      if (this.isAtDocEnd()) return;
      this.idealCol = 0;
      if (this.col === this.buffer.lineLength(this.row)) {
        if (this.row < this.buffer.lineCount()) {
          return this.moveTo(this.row + 1, 0);
        }
      } else {
        return this.moveTo(null, this.col + 1);
      }
    };

    Point.prototype.moveDown = function() {
      if (this.isAtLastLine()) {
        return this.moveToLineEnd();
      } else {
        return this.moveVertical(1);
      }
    };

    Point.prototype.moveUp = function() {
      if (this.row === 0) {
        return this.moveToLineBegin();
      } else {
        return this.moveVertical(-1);
      }
    };

    Point.prototype.moveVertical = function(amount) {
      var limit, newCol, newRow;
      if (!this.idealCol || this.col >= this.idealCol) this.idealCol = this.col;
      newRow = this.row + amount;
      newCol = this.col;
      if (this.idealCol > this.col) newCol = this.idealCol;
      if (this.col > (limit = this.buffer.lineLength(newRow))) newCol = limit;
      return this.moveTo(newRow, newCol);
    };

    Point.prototype.moveToPrevWord = function() {
      var carat, char;
      if (this.isAtDocBegin()) return;
      carat = this.prevLoc();
      char = this.buffer.text(carat.row, carat.col);
      while (!char || !/\w/.test(char)) {
        carat = carat.prevLoc();
        if (carat.isAtDocBegin()) return this.moveToDocBegin();
        char = this.buffer.text(carat.row, carat.col);
      }
      return this.moveTo(this.buffer.wordAt(carat).begin);
    };

    Point.prototype.moveToNextWord = function() {
      var carat, char;
      if (this.isAtDocEnd()) return;
      carat = this.clone();
      char = this.buffer.text(carat.row, carat.col);
      while (!char || !/\w/.test(char)) {
        carat = carat.nextLoc();
        if (carat.isAtDocEnd()) return this.moveToDocEnd();
        if (carat.isAtLineEnd()) carat = carat.nextLoc();
        char = this.buffer.text(carat.row, carat.col);
      }
      carat.moveRight();
      return this.moveTo(this.buffer.wordAt(carat).end);
    };

    Point.prototype.moveToDocBegin = function() {
      return this.moveTo(0, 0);
    };

    Point.prototype.moveToDocEnd = function() {
      var lastRow;
      lastRow = this.buffer.lineCount() - 1;
      return this.moveTo(lastRow, this.buffer.lineLength(lastRow));
    };

    Point.prototype.isAtDocBegin = function() {
      return !this.row && !this.col;
    };

    Point.prototype.isAtDocEnd = function() {
      var lastRow;
      lastRow = this.buffer.lineCount() - 1;
      return this.row === lastRow && this.col === this.buffer.lineLength(lastRow);
    };

    Point.prototype.isAtLineEnd = function() {
      return this.col === this.buffer.lineLength(this.row);
    };

    Point.prototype.isAtLastLine = function() {
      var lastRow;
      lastRow = this.buffer.lineCount() - 1;
      return this.row === lastRow;
    };

    Point.prototype.insert = function(text) {
      return this.moveTo(this.buffer.insert(text, this.row, this.col));
    };

    Point.prototype.overwrite = function(text) {
      return this.moveTo(this.buffer.overwrite(text, this.row, this.col));
    };

    Point.prototype.deleteBack = function() {
      var col, line, row;
      if (this.isAtDocBegin()) return;
      row = this.row, col = this.col;
      this.moveLeft();
      if (col === 0) {
        return this.buffer.joinLines(row - 1);
      } else {
        line = this.buffer.text(row);
        return this.buffer.setLine(row, line.substr(0, col - 1) + line.substr(col));
      }
    };

    Point.prototype.deleteForward = function() {
      var line;
      if (this.isAtDocEnd()) return;
      if (this.isAtLineEnd()) {
        return this.buffer.joinLines(this.row);
      } else {
        line = this.buffer.text(this.row);
        return this.buffer.setLine(this.row, line.substr(0, this.col) + line.substr(this.col + 1));
      }
    };

    Point.prototype.deleteWordBack = function() {
      var colBegin, ptBegin, ptEnd, rowBegin;
      rowBegin = this.row;
      colBegin = this.col;
      this.moveToPrevWord();
      ptBegin = this.buffer.point(this.row, this.col);
      ptEnd = this.buffer.point(rowBegin, colBegin);
      return (new Region(ptBegin, ptEnd))["delete"]();
    };

    Point.prototype.deleteWordForward = function() {
      var colBegin, ptBegin, rowBegin;
      rowBegin = this.row;
      colBegin = this.col;
      this.moveToNextWord();
      ptBegin = this.buffer.point(rowBegin, colBegin);
      return (new Region(ptBegin, this))["delete"]();
    };

    Point.prototype.newLine = function() {
      this.buffer.insertNewLine(this.row, this.col);
      return this.moveTo(this.row + 1, 0);
    };

    return Point;

  })(EventEmitter);

  Region = (function() {

    function Region(begin, end) {
      this.begin = begin;
      this.end = end;
      this.buffer = this.begin.buffer;
    }

    Region.prototype.ordered = function() {
      if (this.begin.isBefore(this.end)) {
        return new Region(this.begin, this.end);
      } else {
        return new Region(this.end, this.begin);
      }
    };

    Region.prototype.isEmpty = function() {
      return this.begin.equals(this.end);
    };

    Region.prototype.text = function() {
      var begin, end, lines, row, _ref, _ref2, _ref3;
      _ref = this.ordered(), begin = _ref.begin, end = _ref.end;
      if (begin.row === end.row) {
        return this.buffer.text(begin.row).substring(begin.col, end.col);
      }
      lines = [];
      lines.push(this.buffer.text(begin.row).substring(begin.col));
      if (end.row - 1 >= begin.row + 1) {
        for (row = _ref2 = begin.row + 1, _ref3 = end.row - 1; _ref2 <= _ref3 ? row <= _ref3 : row >= _ref3; _ref2 <= _ref3 ? row++ : row--) {
          lines.push(this.buffer.text(row));
        }
      }
      lines.push(this.buffer.text(end.row).substring(0, end.col));
      return lines.join("\n");
    };

    Region.prototype.replaceWith = function(text) {
      var afterText, beforeText, begin, delRow, end, lastLine, line, row, _ref, _ref2;
      _ref = this.ordered(), begin = _ref.begin, end = _ref.end;
      line = this.buffer.text(begin.row);
      beforeText = line.substr(0, begin.col);
      if (begin.row === end.row) {
        afterText = line.substr(end.col);
      } else {
        lastLine = this.buffer.text(end.row);
        afterText = lastLine.substr(end.col);
      }
      if (begin.row !== end.row) {
        delRow = begin.row + 1;
        for (row = delRow, _ref2 = end.row; delRow <= _ref2 ? row <= _ref2 : row >= _ref2; delRow <= _ref2 ? row++ : row--) {
          this.buffer.deleteLine(delRow);
        }
      }
      this.buffer.setLine(begin.row, beforeText);
      end.moveTo(this.buffer.insert(text, begin.row, begin.col));
      this.buffer.insert(afterText, end.row, end.col);
    };

    Region.prototype["delete"] = function() {
      return this.replaceWith("");
    };

    Region.prototype.selectRow = function(row) {
      return this.selectRows(row, row);
    };

    Region.prototype.selectRows = function(rowBegin, rowEnd) {
      var _ref;
      if (rowBegin > rowEnd) {
        _ref = [rowEnd, rowBegin], rowBegin = _ref[0], rowEnd = _ref[1];
      }
      this.begin.moveTo(rowBegin, 0);
      return this.end.moveTo(rowEnd, this.buffer.lineLength(rowEnd));
    };

    Region.prototype.shiftLinesUp = function() {
      var begin, end, _ref;
      _ref = this.ordered(), begin = _ref.begin, end = _ref.end;
      if (!this.buffer.shiftLinesUp(begin.row, end.row)) return;
      begin.moveUp();
      return end.moveUp();
    };

    Region.prototype.shiftLinesDown = function() {
      var begin, end, _ref;
      _ref = this.ordered(), begin = _ref.begin, end = _ref.end;
      if (!this.buffer.shiftLinesDown(begin.row, end.row)) return;
      begin.moveDown();
      return end.moveDown();
    };

    Region.prototype.indent = function(tabChars) {
      var begin, end, row, _ref, _ref2, _ref3;
      _ref = this.ordered(), begin = _ref.begin, end = _ref.end;
      for (row = _ref2 = begin.row, _ref3 = end.row; _ref2 <= _ref3 ? row <= _ref3 : row >= _ref3; _ref2 <= _ref3 ? row++ : row--) {
        this.buffer.setLine(row, tabChars + this.buffer.text(row));
      }
      begin.moveTo(null, begin.col + tabChars.length);
      end.moveTo(null, end.col + tabChars.length);
    };

    Region.prototype.outdent = function(tabChars) {
      var begin, beginCol, changed, end, endCol, line, oldLine, re, row, _ref, _ref2, _ref3;
      _ref = this.ordered(), begin = _ref.begin, end = _ref.end;
      re = new RegExp("^" + (tabChars.replace(/(.)/g, "[$1]?")));
      changed = false;
      for (row = _ref2 = begin.row, _ref3 = end.row; _ref2 <= _ref3 ? row <= _ref3 : row >= _ref3; _ref2 <= _ref3 ? row++ : row--) {
        oldLine = this.buffer.text(row);
        line = oldLine.replace(re, "");
        this.buffer.setLine(row, line);
        if (line !== oldLine) changed = true;
      }
      if (!changed) return;
      beginCol = begin.col - tabChars.length;
      if (beginCol < 0) beginCol = 0;
      endCol = end.col - tabChars.length;
      if (endCol < 0) endCol = 0;
      begin.moveTo(null, beginCol);
      return end.moveTo(null, endCol);
    };

    return Region;

  })();

  module.exports = {
    Buffer: Buffer,
    Point: Point,
    Region: Region
  };

}).call(this);
